function [] = Ex_28p2_002_Laplaces_method()
% ---------------------------------------------------------------------------
% Exercise 28.2 in MacKay's book; by Matthew Juniper 2020
% ---------------------------------------------------------------------------
% PRIOR READING
%  Information Theory, Inference, and Learning Algorithms, by David J.C.
%  MacKay, Cambridge University Press (v7.2 2005),
%  https://www.inference.org.uk/itprnn/book.pdf 
% 
% Essential reading:
%  Chapters  1 -  3; 
%  Chapters 27 - 28.
%
% Useful reading:
%  Chapters 21, 22, 24.
%  Exercise 28.1
% ---------------------------------------------------------------------------

% ---------------------------------------------------------------------------
% QUESTION
% ---------------------------------------------------------------------------
% We are given data, D, and asked to evaluate the evidence (a.k.a. marginal
% likelihood) for two models, H1(w) and H2(w), given the data, D, where
% w are the models' parameters. In other words:
% 
%   Calculate the evidences, P(D|H1) and P(D|H2), where:
% 
%   P(D|Hi) = int{ P(D|w,Hi) * P(w|Hi) } dw  (Eq. 28.7) 
% 
%   D is [x] = [ -8 , -2 ,  6 ]
%        [t]   [  8 , 10 , 11 ]
%
%   The measurement, t, is y + Normal(0,1)
% 
%   Model H1 is y = w0
%   Model H2 is y = w0 + w1*x
% 
%   Prior w0 is Normal(0,1) for both models
%   Prior w1 is Normal(0,1)

% ---------------------------------------------------------------------------
% SOLUTION WITH LAPLACE'S METHOD (Chapter 27)
% ---------------------------------------------------------------------------
% The symbolic algebra method is relatively slow and does not scale well.
% We will answer the same question with Laplace's method because it is
% quicker and scales better. 
%
% In Laplace's method, we find the peak of P(D|w,Hi) and Taylor expand to
% second order around this peak. We then approximate P(D|w,Hi) by a
% Gaussian and we approximate the normalizing constant of P(D|w,Hi) by the
% normalizing constant of this Gaussian. 
%
% In this question, P(D|w,Hi) is Gaussian, so the solution with Laplace's
% method will equal that found with symbolic algebra. This provides a useful
% check.
%
% The algebra is written for model H2; model H1 is a special case of H2.
% The algebra includes variables for expected values and variances so that
% they can be changed later. The measurement noise remains Normal(0,1).

% Enter the data from Exercise 28.2 in MacKay's book
x = [ -8 , -2 ,  6 ];
t = [  8 , 10 , 11 ];
% Enter the prior expected values and variances of w0 and w1
w0_E = 0.0; w0_V = 1.0;
w1_E = 0.0; w1_V = 1.0;

% Start the clock
tic

% ---------------------------------------------------------------------------
% Prior, Likelihood, and Posterior for Model H2
% ---------------------------------------------------------------------------
%
% 1. Consider Model H2
%
%   y = w0 + w1*x
%
% 2a. The Prior distribution for parameter w0 is:
%
%   p(w0|H2) = Normal(w0_E,w0_V)
%
%                     1              {   1 (w0_E - w0)^2 }
%   p(w0|H2) = --------------- * exp { - - ------------- }
%              sqrt(2*pi*w0_V)       {   2     w0_V      }
%
% 2b. The Prior distribution for parameter w1 is:
%
%   p(w1|H2) = Normal(w1_E,1)
%
%                     1              {   1 (w1_E - w1)^2 }
%   p(w1|H2) = --------------- * exp { - - ------------- }
%              sqrt(2*pi*w1_V)       {   2     w1_V      }
%
% 3. The Posterior distribution for w0,w1 is given by:
% 
%                   P(D|w0,w1,H2) * P(w0|H2) * P(w1|H2)
%   P(w0,w1|D,H2) = -----------------------------------       [1]
%                               P(D|H2)
%
% 4. The measurement, t, is y + epsilon, where epsilon is Normal(0,1):
%
%   t = y + epsilon
% 
%                                    1            {   1 (t-y(w0,w1,H2))^2 }
%   P(t|w0,w1,H2) = P(epsilon) = ---------- * exp { - - ----------------- }
%                                sqrt(2*pi)       {   2         1         }
% 
%   where y(w0,w1,H2) = w0 + w1*x for model H2
%
% 5. the likelihood is:
% 
%   P(D|w0,w1,H2) = P(t1|w0,w1,H2) * P(t2|w0,w1,H2) * P(t3|w0,w1,H2)
%
%                       1             {   1 (t1-w0-w1*x1)^2 + (t2-w0-w1*x2)^2 + (t3-w0-w1*x3)^2 }      . 
%              = -------------- * exp { - - --------------------------------------------------- } [2]  .
%                (sqrt(2*pi))^3       {   2                          1                          }      . 
%
% Note that the measurement variance is hard-wired to 1. User-defined
% measurement variance can be introduced by replacing 1 with t_V in the
% denominator of the argument and dividing the pre-exponential factor by
% sqrt(t_V). That change would then follow through the algebra below.
%
% 6. The Posterior distribution is:
%
%   P(w0,w1|D,H2)  =  P(D|w0,w1,H2) * P(w0|H2) * P(w1|H2)

% ---------------------------------------------------------------------------
% Find (w0,w1) that maximizes the Posterior distribution P(w0,w1|D,H2)
% ---------------------------------------------------------------------------
%
% We don't need to know the denominator of [1], P(D|H2), for this. We just
% maximize the numerator of the posterior distribution above: 
%
% Find w0,w1 that maximizes P(D|w0,w1,H2) * P(w0|H2) * P(w1|H2), which occurs when:
%
%       d   ( (t1-w0-w1*x1)^2 + (t2-w0-w1*x2)^2 + (t3-w0-w1*x3)^2   (w0_E-w0)^2   (w1_E-w1)^2 )
% 0 = ----- ( --------------------------------------------------- + ----------- + ----------- )
%     d(w0) (                          1                               w0_V           w1_V    )
%
% and
%
%       d   ( (t1-w0-w1*x1)^2 + (t2-w0-w1*x2)^2 + (t3-w0-w1*x3)^2   (w0_E-w0)^2   (w1_E-w1)^2 )
% 0 = ----- ( --------------------------------------------------- + ----------- + ----------- )
%     d(w1) (                          1                               w0_V           w1_V    )
%
% The first of these gives:
%
% 0 = -(t1-w0-w1*x1) - (t2-w0-w1*x2) - (t3-w0-w1*x3) - (w0_E-w0)/w0_V
% 0 = (3+1/W0_V)*w0 - (t1-w1*x1) - (t2-w1*x2) - (t3-w1*x3) - w0_E/w0_V
% 0 = (3+1/W0_V)*w0 + w1*(x1+x2+x3) - (t1+t2+t3) - w0_E/w0_V                       [1a]
%
% The second of these gives
%
% 0 = - 2*x1*(t1-w0-w1*x1) - 2*x2*(t2-w0-w1*x2) - 2*x3*(t3-w0-w1*x3) - 2*(w1_E-w1)/w1_V  
% 0 = - x1*(t1-w0-w1*x1) - x2*(t2-w0-w1*x2) - x3*(t3-w0-w1*x3) - (w1_E/w1_V-w1/w1_V)  
% 0 = w0(x1+x2+x3) + w1*(x1^2+x2^2+x3^2+1/w1_V) - (t1*x1+t2*x2+t3*x3) - w1_E/w1_V     [1b]
%
% Put into matrix form:
%  M * [w0] = [r0]          [1a]
%      [w1]   [r1]          [1b]
% LHS:
M_H2 = [ length(t)+1/w0_V      , sum(x)              ; ... % [LHS of 1a]
         sum(x)                , sum(x.^2) + 1/w1_V ];     % [LHS of 1b]
% RHS:
r_H2 = [ sum(t) + w0_E/w0_V     ; ... % [RHS of 1a]
         sum(t.*x) + w1_E/w1_V ];     % [RHS of 1b]
% Solve for w0 and w1 using linear algebra w = inv(M) * r
w = M_H2\r_H2;
% Extract w0 and w1 from w
w0_MAP_H2 = w(1); w1_MAP_H2 = w(2);

% ---------------------------------------------------------------------------
% Prior, Likelihood, and Posterior for Model H1
% ---------------------------------------------------------------------------
%
%   y = w0
%
% Find w0 that maximizes P(D|w0,H1) * P(w0|H1), which occurs when:
%
%       d   ( (t1-w0)^2 + (t2-w0)^2 + (t3-w0)^2   (w0_E-w0)^2 )
% 0 = ----- ( --------------------------------- + ----------- )
%     d(w0) (                       1                w0_V     )
w0_MAP_H1 = (sum(t) + w0_E/w0_V) / (length(t)+1/w0_V);

% ---------------------------------------------------------------------------
% Perform Taylor expansion of log(f[w0,w1]) around the MAP point for H2
% ---------------------------------------------------------------------------
%
% The numerator of the posterior f:= P(w0,w1|D,H2) in [1] is:
%
%                   1                    {   1 ( (t1-w0-w1*x1)^2 + (t2-w0-w1*x2)^2 + (t3-w0-w1*x3)^2   (w0_E-w0)^2   (w1_E-w1)^2 ) }  . 
% f = ---------------------------- * exp { - - ( --------------------------------------------------- + ----------- + ----------- ) }  .
%     sqrt((2*pi)^5 * w0_V * w1_V)       {   2 (                       1                                  w0_V           w1_V    ) }  . 
%
% Take the log of this:
%
% g = log(f) = -2.5*log(2*pi) - log(w0_V) - log(w1_V) - 0.5*(t1-w0-w1*x1)^2 + 0.5*(t2-w0-w1*x2)^2 + 0.5*(t3-w0-w1*x3)^2 + 0.5*(w0_E-w0)^2/w0_V + 0.5*(w1_E-w1)^2/w1_V 
%
% Perform a second order Taylor expansion on g:
%                                                                                          <------- Matrix A ------->  
%                        dg                dg                                              [ d2g/dw0dw0  d2g/dw0dw1 ]   [ delta(w0) ]                                     . 
% delta(g) = g[w0,w1]  + --- * delta(w0) + --- * delta(w1)  +  [ delta(w0) , delta(w1) ] * [                        ] * [           ]  +  [third and higher order terms]  . 
%                        dw0               dw1                                             [ d2g/dw1dw0  d2g/dw1dw1 ]   [ delta(w0) ]                                     .
%                        <------ first order terms ------>     <--------------------- second order terms --------------------------->                                     .
%
% dg
% --- = - (t1-w0-w1*x1) - (t2-w0-w1*x2) - (t3-w0-w1*x3) - (w0_E-w0)/w0_V
% dw0
%
%  d2g     
% ----- = length(t) + 1/w0_V
% dw0^2
%
% dg
% --- = - x1*(t1-w0-w1*x1) - x2*(t2-w0-w1*x2) - x3*(t3-w0-w1*x3) - (w1_E-w1)/w1_V
% dw1
%
%  d2g     
% ----- = x1^2 + x2^2 + x3^2 + 1/w1_V 
% dw1^2
%
%  d2g
% ------ = x1 + x2 + x3 
% dw0dw1
% 
% Note that, in this case, there are no third and higher order terms to the
% Taylor expansion of g. This is because g is quadratic, which is because
% the distributions that make up f are Gaussian. This is not true in general.  
%
% At the MAP point, dg/dw0 and dg/dw1 are zero (by definition of the MAP
% point). Therefore the first order terms in the Taylor expansion disappear
% and only the second order terms remain. 
%
% Around the MAP point, the posterior, f, takes the form:
%
%                 {   1                   [dw0] }
% f = f_max * exp { - - [dw0 , dw1] * A * [   ] }
%                 {   2                   [dw1] }
%
% where 
%   f_max is the peak value of f, occurring at (w0_max,w1_max)
%   dw0 = w0 - w0_max 
%   dw1 = w1 - w1_max 
% and
%
% A is the matrix labelled 'matrix A' above (repeated here for convenience):
%
%       [ d2g/dw0dw0  d2g/dw0dw1 ]
%   A = [                        ]
%       [ d2g/dw1dw0  d2g/dw1dw1 ]

% Perform the 2nd order Taylor expansion for H2
A_H2 = [ length(t) + 1/w0_V ,  sum(x)             ; ...
         sum(x)             ,  sum(x.^2) + 1/w1_V ];

% Perform the 2nd order Taylor expansion for H1 (by inspection)
A_H1 =   length(t) + 1/w0_V;

% ---------------------------------------------------------------------------
% Use Laplace's approximation to calculate the evidence for H2; from MacKay p350 equ (28.10)
% ---------------------------------------------------------------------------
%
% The integral of the posterior f, over the whole of parameter space (w0,w1) is approximated by: 
%
%                                              Matrix A written above 
%           <likelihood->   <--prior-->        |
% P(D|H2) ~ P(D|w_max,H2) * P(w_max|H2) * [det(A/2/pi)]^(-1/2)
%           <--- what I call f_max --->
%
%     f_max_H2 = f_max_H2 * (1/sqrt(2*pi) * exp( -(t(nn)-(w0_max_H2 + w1_max_H2*x(nn))).^2 ));
%
% logf_max is the value of logf at the peak, i.e. where w_max = [w0_max_H2,w1_max_H2]
logf_max_H2 = 0;
% Calculate the contribution of each datapoint to the log likelihood log{P(D|w_max,H2)}
for nn = 1:length(t)
    logf_max_H2 = logf_max_H2 - 0.5*log(2*pi) - 0.5 * ( t(nn) - ( w0_MAP_H2 + w1_MAP_H2*x(nn) ) ).^2 ;
end
% Calculate the contribution of the prior to logf_max
logf_max_H2 = logf_max_H2 - 0.5*log(2*pi*w0_V) - 0.5*log(2*pi*w1_V) - 0.5*(w0_E-w0_MAP_H2).^2/w0_V - 0.5*(w1_E-w1_MAP_H2).^2/w1_V;
% Estimate the logevidence with Laplace's method
log_P_D_H2 = logf_max_H2 - 0.5*log(det(A_H2/2/pi)); 

% ---------------------------------------------------------------------------
% Use Laplace's approximation to calculate the evidence for H1
% ---------------------------------------------------------------------------
%
logf_max_H1 = 0;
% Calculate the contribution of each datapoint to the log likelihood log{P(D|w_max,H1)}
for nn = 1:length(t)
    logf_max_H1 = logf_max_H1 - 0.5*log(2*pi) - 0.5 * ( t(nn) - w0_MAP_H1 ).^2 ;
end
% Calculate the contribution of the prior to logf_max
logf_max_H1 = logf_max_H1 - 0.5*log(2*pi*w0_V) - 0.5*(w0_E-w0_MAP_H1)^2/w0_V;
% Estimate the logevidence with Laplace's method
log_P_D_H1 = logf_max_H1 - 0.5*log(A_H1/2/pi);

% Stop the clock
time_taken = toc;

fprintf('------------------------------------------------ \n')
fprintf(' log(evidence) calculated with Laplace''s method \n')
fprintf('------------------------------------------------ \n')
fprintf('log{ P(D|H1) } = %6.4f \n',log_P_D_H1)
fprintf('log{ P(D|H2) } = %6.4f \n',log_P_D_H2)
fprintf('------------------------------------------------ \n')
fprintf('P(D|H1) / P(D|H2) = %6.4f \n',exp(log_P_D_H1 - log_P_D_H2))
fprintf('------------------------------------------------ \n')
fprintf(' MAP point calculated with hand-derived algebra  \n')
fprintf('------------------------------------------------ \n')
fprintf('max{ P(D|H1) } occurs at w = ( %6.4f )   \n',w0_MAP_H1)
fprintf('max{ P(D|H2) } occurs at w = ( %6.4f , %6.4f ) \n',w0_MAP_H2,w1_MAP_H2)
fprintf('------------------------------------------------ \n')
fprintf('time taken for Laplace''s method with hand-derived algebra = %6.4f secs \n',time_taken)

end
